// Search and replace a hex-string from a file.
// Length can be different. 
//
// James 2022 - converted from hexpatch.hpp, which in turn is converted
// from magisk source. Magisk is GNU GPL v3, so this one is GNU GPL v3 too.
//
// Compile as:
// dynamic:          gcc -Wl,-s -o hexpatch hexpatch.c
// static with glic: gcc -static -Wl,-s -o hexpatch hexpatch.c
// static with musl: musl-gcc -static -Wl,-s -o hexpatch hexpatch.c -I /usr/musl/kernel/include

#define _GNU_SOURCE
#include <unistd.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/fs.h>
#include <alloca.h>
/*
#include <sys/fcntl.h>
*/

#define FALSE 0
#define TRUE (!FALSE)

struct data_s {
	char *buf;
	size_t sz;
};

char *mmap_data(const char *name, int rw, struct data_s *data) {
	data->buf = NULL;
	data->sz = 0;
	
    int fd = open(name, (rw ? O_RDWR : O_RDONLY) | O_CLOEXEC);
    if (fd < 0)
        return 0;
    struct stat st;
    if (fstat(fd, &st))
        return 0;
    if (S_ISBLK(st.st_mode)) {
        uint64_t size;
        ioctl(fd, BLKGETSIZE64, &size);
        data->sz = size;
    } else {
        data->sz = st.st_size;
    }
    void *b = data->sz > 0
            ? mmap(NULL, data->sz, PROT_READ | PROT_WRITE, rw ? MAP_SHARED : MAP_PRIVATE, fd, 0)
            : NULL;
    close(fd);
    return data->buf = (char *)b;
}

static void hex2byte(const char *hex, uint8_t *buf) {
    char high, low;
    for (int i = 0, length = strlen(hex); i < length; i += 2) {
        high = toupper(hex[i]) - '0';
        low = toupper(hex[i + 1]) - '0';
        buf[i / 2] = ((high > 9 ? high - 7 : high) << 4) + (low > 9 ? low - 7 : low);
    }
}

int hexpatch(const char *file, const char *from, const char *to) {
    int patched = 1;
	struct data_s data;

	// sanity check
	int pattern_len = strlen(from)/2;
	int patch_len = strlen(to)/2;
	
	if (pattern_len & 1) return patched; // must be even
	if (patch_len & 1) return patched; // must be even
	
	if (mmap_data(file, TRUE, &data) != NULL) {

		char *pattern = alloca(pattern_len);
		char *patch = alloca(patch_len);

		hex2byte(from, pattern);
		hex2byte(to, patch);

		char * const end = data.buf + data.sz;
		for (char *curr = data.buf; curr < end; curr += pattern_len) {
			curr = (char *) (memmem(curr, end - curr, pattern, pattern_len));
			if (curr == NULL)
				return patched;
			fprintf(stderr, "Patch @ %08X [%s] -> [%s]\n", (unsigned)(curr - data.buf), from, to);
			memset(curr, 0, pattern_len);
			memcpy(curr, patch, patch_len);
			patched = 0;
		}
	}

    return patched;
}

int main(int argc, char **argv) {
    const char *infile = argv[1];
    const char *from = argv[2];
    const char *to = argv[3];
    if (argc < 3) {
        fprintf(stderr, "Usage:\n\t./hexpatch [file] [from_x] [to_X]\n");
        return 1;
    }

    if (access(infile, F_OK) !=0) {
        fprintf(stderr, "Error : File does not exist !\n");
        return 1;
    }
    return hexpatch(infile, from, to);
}
